#!/usr/bin/env python3
#
# Usage: ./gen.py basenames
# where basenames are like getrf, gesv, posv, ...
#
# Generates wrapper in ../gen, copies to ../src, and attempts to compile.

from __future__ import print_function

import sys
import os

pwd = os.getcwd()

def run( cmd ):
    print( cmd )
    os.system( cmd )
# end

for arg in sys.argv[1:]:
    try:
        gen = '../gen/' + arg + '.cc'
        src = '../src/' + arg + '.cc'
        if (os.path.exists( src )):
            print( src + ' already exists; skipping' )
            continue
        run( './wrapper_gen.py ' + arg )
        run( 'cp ' + gen + ' ' + src )
        os.chdir( '../src' )
        run( 'make ' + arg + '.o' )
    except:
        pass

    os.chdir( pwd )
# end



#!/usr/bin/env python3
#
# Attempts to find equations in Doxygen code and replace them with $...$.
# For instance:
#     /// solve A * X = B^H for matrices A, X, B.
# becomes
#     /// solve $A X = B^H$ for matrices A, X, B.
# However, patterns are heuristic and incomplete, so mistakes occur.
# Use with caution.

from __future__ import print_function

import sys
import re
import os

debug = False

operators = {
    '<=': r'\le',
    '>=': r'\ge',
}

src = '../gen'
assert( os.path.exists( src ))

dst = '../eqn'
if (not os.path.exists( dst )):
    os.mkdir( dst )

def equation( line, subexpr=False ):
    if (debug):
        print( 'equation:', line + '; subexpr:', subexpr )

    if (subexpr):
        eqn = ''
    else:
        eqn = r'$'
    begin = True
    while (line):
        # no space after $ at beginning;
        # after that, add a space before most tokens (except ^T and punctuation)
        if (begin):
            begin = False
            space = ''
        else:
            space = ' '

        if (debug):
            print( 'eqn: <' + eqn + '> line: <' + line + '>' )

        # space
        s = re.search( r'^( +)(.*)', line )
        if (s):
            line = s.group(2)
            continue

        # variable
        s = re.search( r'^([a-zA-Z]|VT)(\d*)\b(.*)', line )
        if (s):
            eqn += space + s.group(1)
            if (s.group(2)):
                eqn += '_{' + s.group(2) + '}'
            line = s.group(3)
            continue

        # number
        s = re.search( r'^(\d+)\b(.*)', line )
        if (s):
            eqn += space + s.group(1)
            line = s.group(2)
            continue

        # binary operator * becomes implicit
        s = re.search( r'^(\*)(.*)', line )
        if (s):
            line = s.group(2)
            continue

        # explicit binary operators <= >=
        s = re.search( r'^(<=|>=)(.*)', line )
        if (s):
            eqn += space + operators[ s.group(1) ]
            line = s.group(2)
            continue

        # explicit binary operators = < > + - /
        s = re.search( r'^([=<>\+\-\/])(.*)', line )
        if (s):
            eqn += space + s.group(1)
            line = s.group(2)
            continue

        # exponents, transpose
        s = re.search( r'^(\^[HT2-9])\b(.*)', line )
        if (s):
            # don't print space
            eqn += s.group(1)
            line = s.group(2)
            continue

        # Greek, functions
        s = re.search( r'^(alpha|beta|lambda|Lambda|mu|sigma|Sigma|tau|sqrt|min|max)\b(.*)', line )
        if (s):
            eqn += space + '\\' + s.group(1)
            line = s.group(2)
            continue

        # capitalized Greek, functions
        s = re.search( r'^(MIN|MAX)\b(.*)', line )
        if (s):
            eqn += space + '\\' + s.group(1).lower()
            line = s.group(2)
            continue

        # norm
        # todo: use \lVert?
        s = re.search( r'^norm\((.*)', line )
        if (s):
            (expr, line) = equation( s.group(1), True )
            eqn += space + r'|| ' + expr + ' ||'
            continue

        # abs
        # todo: use \lvert?
        s = re.search( r'^abs\((.*)', line )
        if (s):
            (expr, line) = equation( s.group(1), True )
            eqn += space + r'| ' + expr + ' |'
            continue

        # inverse
        s = re.search( r'^inv\((.*)', line )
        if (s):
            (expr, line) = equation( s.group(1), True )
            if (re.search( '^\w+$', expr )):
                # safe to skip parens
                eqn += space + expr + '^{-1}'
            else:
                eqn += space + '(' + expr + ')^{-1}'
            continue

        # non-Latex functions
        s = re.search( r'^(diag)\((.*)', line )
        if (s):
            (expr, line) = equation( s.group(2), True )
            eqn += space + r'\; \text{' + s.group(1) + '}(' + expr + ') \; '
            continue

        # open parens
        s = re.search( r'^\((.*)', line )
        if (s):
            (expr, line) = equation( s.group(1), True )
            if (expr):
                eqn += space + '(' + expr + ')'
            else:
                # sub-expr didn't parse, so put ( back on line
                line = '(' + line
                break # done
            continue

        # close parens
        if (subexpr):
            s = re.search( r'^\)(.*)', line )
            if (s):
                # finished subexpression; don't include ) in eqn
                line = s.group(1)
                if (debug):
                    print( 'finish subexpr:', eqn, '\nline:', line )
                break

        # ellipsis
        s = re.search( r'^\. ?\. ?\.(.*)', line )
        if (s):
            eqn += space + r'\dots'
            line = s.group(1)
            continue

        # punctuation
        ##s = re.search( r'^([.,;:])(.*)', line )
        s = re.search( r'^([,])(.*)', line )
        if (s):
            # don't print space
            eqn += s.group(1)
            line = s.group(2)
            continue

        # otherwise done
        break
    # end
    if (not subexpr):
        eqn += '$'
        # add space after $,
        # unless there's punctuation or already space such as \n in line.
        s = re.search( r'^[^.,;:\s]', line )
        if (s):
            eqn += ' '
    return (eqn, line)
# end

def process( arg ):
    filename = arg + '.cc'
    f_out = os.path.join( dst, filename )
    f_in  = os.path.join( src, filename )
    print( 'reading ' + f_in + '\nwriting ' + f_out )
    f_out = open( f_out, 'w' )
    f_in  = open( f_in )
    for line in f_in:
        s = re.search( r'^///', line )
        if (s):
            rest = line.rstrip()
            line = ''
            while (rest):
                #                                variable        Greek                                            operator          | functions
                s = re.search( r'^(.*?)\b((?: (?:[A-Z]\d* | VT | alpha | beta | lambda | mu | sigma | tau) \s* (?:=|\^|\+|\-|\*|\/) | norm\( | inv\( | diag\( | sqrt\( ).*)',
                               rest, re.X )
                if (s):
                    line += s.group(1)
                    (eqn, rest) = equation( s.group(2) )
                    if (debug):
                        print( 'eqn: ' + eqn + ', rest: ' + rest )
                    eqn = re.sub( r',\$( *)$', r'$,\1', eqn )  # move trailing punctuation (only commas) outside
                    line += eqn
                else:
                    break
                # end
            # end
            line += rest + '\n'
            # if the entire line is an indented equation, make it \[ ... \]
            # and put punctuation inside.
            line = re.sub( r'^///     \$(.*)\$([.,;:]?)$', r'///     \[ \1\2 \]', line )
        # end
        print( line, file=f_out, end='' )
    # end
# end

for arg in sys.argv[1:]:
    if (arg == '-d'):
        debug = True
    else:
        process( arg )
# end



#!/usr/bin/env python3

'''
Generate C prototypes for the LAPACK Fortran functions by parsing
doxygen comments in LAPACK's SRC *.f files. Looks for routines in
LAPACK's SRC, INSTALL, and MATGEN directories. (Because MATGEN is the
last directory checked, error messages show that file not found.)

Example usage:

    # Assumes $LAPACKDIR is set
    lapackpp> echo $LAPACKDIR
    /Users/mgates/Documents/lapack

    lapackpp> ./tools/header_gen.py {s,d,c,z}posv {s,d}ormqr {c,z}unmqr
    generating gen/fortran.h

    lapackpp> ./tools/header_gen.py all
    reading include/lapack/fortran.h to re-generate all prototypes
    generating gen/fortran.h
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/slassq.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/dlassq.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/classq.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/zlassq.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/slartg.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/dlartg.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/clartg.f
    // skipping, file not found: /Users/mgates/Documents/lapack/TESTING/MATGEN/zlartg.f
'''

from __future__ import print_function

import sys
import re
import os

lapack = os.environ['LAPACKDIR']

if (len( sys.argv ) == 1):
    print( 'Usage:', sys.argv[0], '[routine ...]\n' +
           '      ', sys.argv[0], 'all\n' +
           'See script for more details' )
    exit(1)
# end

if (sys.argv[1] == 'all'):
    #header = lapack + '/LAPACKE/include/lapack.h'
    header = 'include/lapack/fortran.h'
    print( 'reading', header, 'to re-generate all prototypes' )
    funcs = []
    f = open( header )
    for line in f:
        s = re.search( r' +LAPACK_(\w+) +LAPACK_GLOBAL\( *\1', line )
        #s = re.search( r'(?:void|float|double) +LAPACK_GLOBAL\((\w+),', line )
        if (s):
            funcs.append( s.group(1) )
    # end
else:
    funcs = sys.argv[1:]
# end

output = 'gen/fortran.h'
print( 'generating', output )
output = open( output, 'w' )

intent = None
var    = None
vtype  = None
dtype  = None

typemap = {
    'character'        : 'char',
    'integer'          : 'lapack_int',
    'real'             : 'float',
    'double precision' : 'double',
    'complex'          : 'lapack_complex_float',
    'complex*16'       : 'lapack_complex_double',
    'logical'          : 'lapack_logical',

    'logical function of two real arguments'               : 'LAPACK_S_SELECT2',
    'logical function of three real arguments'             : 'LAPACK_S_SELECT3',

    'logical function of two double precision arguments'   : 'LAPACK_D_SELECT2',
    'logical function of three double precision arguments' : 'LAPACK_D_SELECT3',

    'logical function of one complex argument'     : 'LAPACK_C_SELECT1',
    'logical function of two complex arguments'    : 'LAPACK_C_SELECT2',

    'logical function of one complex*16 argument'  : 'LAPACK_Z_SELECT1',
    'logical function of two complex*16 arguments' : 'LAPACK_Z_SELECT2',
}

for func in funcs:
    s = re.search( '\.f', func )
    if (s):
        filename = func
    else:
        for subdir in ('SRC', 'INSTALL', 'TESTING/MATGEN'):
            filename = lapack + '/' + subdir + '/' + func + '.f'
            if (os.path.exists( filename )):
                break
            # end
        # end
    # end

    if (os.path.exists( filename )):
        #print( 'read', filename )
        f = open( filename )
        args = []
        retval = 'void'
        for line in f:
            s = re.search( r'^\* +\b(.*)\b +FUNCTION +(\w+)', line )  # \( *(.*) *\)
            if (s):
                retval = typemap[ s.group(1).lower() ]
                continue
            # end

            s = re.search( r'^\*> *\\param\[(in|out|in,out|inout)\] +(\w+)', line )
            if (s):
                intent = s.group(1).lower()
                var    = s.group(2)
                vtype  = var  # vtype used as flag to say we have a variable and are looking for its type
                continue
            # end

            if (vtype):
                s = (re.search( '^\*> *' + vtype + ' +is +(CHARACTER|INTEGER|REAL|DOUBLE PRECISION|COMPLEX\*16|COMPLEX|LOGICAL)( array)?', line ) or
                     re.search( '^\*> *\(workspace\) +(CHARACTER|INTEGER|REAL|DOUBLE PRECISION|COMPLEX\*16|COMPLEX|LOGICAL)( array)?', line ) or
                     re.search( '^\*> *' + vtype + ' +is a (LOGICAL FUNCTION of (?:one|two|three) (?:REAL|DOUBLE PRECISION|COMPLEX\*16|COMPLEX) arguments?)( array)?', line ))
                if (s):
                    dtype = s.group(1).lower()
                    dtype = typemap[dtype]
                    array = s.group(2)
                    vtype = None
                    # leave arrays uppercase, except work and pivots
                    if (array and var not in ('WORK', 'RWORK', 'IWORK', 'IPIV', 'JPIV', 'PIV')):
                        pass #lvar = var
                    else:
                        var = var.lower()
                    if (intent == 'in'):
                        args.append( dtype + ' const* ' + var )
                    else:
                        args.append( dtype + '* ' + var )
                    continue
                # end
            # end
        # end
        print( '#define LAPACK_' + func + ' LAPACK_GLOBAL( ' + func + ', '
               + func.upper() + ' )', file=output )
        print( retval + ' LAPACK_' + func + '(\n    ' + ', '.join( args )
               + '\n);\n', file=output )
    else:
        print( '// skipping, file not found:', filename )
        print( '// skipping, file not found:', filename, file=output )
        print( '// #define LAPACK_' + func + ' LAPACK_GLOBAL(' + func + ','
               + func.upper() + ')', file=output )
        print( '// void LAPACK_' + func + '( ... );\n', file=output )
    # end
# end

output.close()



#!/usr/bin/env python3
#
# Copyright (c) 2017-2023, University of Tennessee. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# This program is free software: you can redistribute it and/or modify it under
# the terms of the BSD 3-Clause license. See the accompanying LICENSE file.
#
# Usage: python3 configure.py [--interactive]

from __future__ import print_function

import sys
import re
import config
from   config import Error, font, print_warn, print_header
import config.lapack

#-------------------------------------------------------------------------------
# header

print( '-'*80 + '\n' +
font.bold( font.blue( '                              Welcome to LAPACK++.' ) ) +
'''

By default, configure will automatically choose the first valid value it finds
for each option. You can set it to interactive to find all possible values and
give you a choice:
    ''' + font.blue( 'make config interactive=1' ) + '''

If you have multiple compilers, we suggest specifying your desired compiler by
setting CXX, as the automated search may prefer a different compiler.

For options, see the `INSTALL.md` file.

Configure assumes environment variables CPATH, LIBRARY_PATH, and LD_LIBRARY_PATH
are set so your compiler can find libraries. See INSTALL.md for more details.
''' + '-'*80 )

#-------------------------------------------------------------------------------
def main():
    config.init( namespace='LAPACK', prefix='/opt/slate' )
    config.prog_cxx()

    print_header( 'C++ compiler flags' )
    # Pick highest level supported. oneAPI needs C++17.
    # Crusher had issue with -std=c++20 (2022-07).
    config.prog_cxx_flag(
        ['-std=c++17', '-std=c++14', '-std=c++11'])
    config.prog_cxx_flag( '-O2' )
    config.prog_cxx_flag( '-MMD' )
    config.prog_cxx_flag( '-Wall' )
    config.prog_cxx_flag( '-Wno-unused-local-typedefs' )
    config.prog_cxx_flag( '-Wno-unused-function' )
   #config.prog_cxx_flag( '-pedantic',  # todo: conflict with ROCm 3.9.0
   #config.prog_cxx_flag( '-Wshadow',   # todo: conflict with ROCm 3.9.0
   #config.prog_cxx_flag( '-Wmissing-declarations' )
   #config.prog_cxx_flag( '-Wconversion' )
   #config.prog_cxx_flag( '-Werror' )

    config.openmp()

    config.lapack.blas()
    print()
    config.lapack.blas_float_return()
    config.lapack.blas_complex_return()
    config.lapack.vendor_version()

    # Must test mkl_version before cblas and lapacke, to define HAVE_MKL.
    try:
        config.lapack.cblas()
    except Error:
        print_warn( 'LAPACK++ needs CBLAS for testers.' )

    config.lapack.lapack()
    config.lapack.lapack_version()

    # XBLAS and Matgen are optional
    try:
        config.lapack.lapack_xblas()
    except Error:
        print_warn( 'LAPACK++ will exclude wrappers for XBLAS.' )

    try:
        config.lapack.lapack_matgen()
    except Error:
        print_warn( 'LAPACK++ will exclude wrappers for matgen.' )

    try:
        config.lapack.lapacke()
    except Error:
        print_warn( 'LAPACK++ needs LAPACKE for testers.' )

    config.gpu_blas()

    blaspp = config.get_package(
        'BLAS++',
        ['../blaspp', './blaspp'],
        'include/blas.hh',
        'https://github.com/icl-utk-edu/blaspp',
        'https://github.com/icl-utk-edu/blaspp/tarball/master',
        'blaspp.tar.gz' )
    if (not blaspp):
        raise Exception( 'LAPACK++ requires BLAS++.' )

    testsweeper = config.get_package(
        'TestSweeper',
        ['../testsweeper', blaspp + '/testsweeper', './testsweeper'],
        'testsweeper.hh',
        'https://github.com/icl-utk-edu/testsweeper',
        'https://github.com/icl-utk-edu/testsweeper/tarball/master',
        'testsweeper.tar.gz' )
    if (not testsweeper):
        print_warn( 'LAPACK++ needs TestSweeper for testers.' )

    config.extract_defines_from_flags( 'CXXFLAGS', 'lapackpp_header_defines' )
    config.output_files( ['make.inc', 'include/lapack/defines.h'] )
    print( 'log in config/log.txt' )

    print( '-'*80 )
# end

#-------------------------------------------------------------------------------
try:
    main()
except Error as ex:
    print_warn( 'A fatal error occurred. ' + str(ex) +
                '\nLAPACK++ could not be configured. Log in config/log.txt' )
    exit(1)



#!/usr/bin/env python3
#
# Copyright (c) 2017-2023, University of Tennessee. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# This program is free software: you can redistribute it and/or modify it under
# the terms of the BSD 3-Clause license. See the accompanying LICENSE file.
#
# Example usage:
# help
#     ./run_tests.py -h
#
# run everything with default sizes
# output is redirected; summary information is printed on stderr
#     ./run_tests.py > output.txt
#
# run LU (gesv, getrf, getri, ...), Cholesky (posv, potrf, potri, ...)
# with single, double and default sizes
#     ./run_tests.py --lu --chol --type s,d
#
# run getrf, potrf with small, medium sizes
#     ./run_tests.py -s -m getrf potrf

from __future__ import print_function

import sys
import os
import re
import argparse
import subprocess
import xml.etree.ElementTree as ET
import io
import time

# ------------------------------------------------------------------------------
# command line arguments
parser = argparse.ArgumentParser()

group_test = parser.add_argument_group( 'test' )
group_test.add_argument( '-t', '--test', action='store',
    help='test command to run, e.g., --test "mpirun -np 4 ./test"; default "%(default)s"',
    default='./tester' )
group_test.add_argument( '--xml', help='generate report.xml for jenkins' )
group_test.add_argument( '--dry-run', action='store_true', help='print commands, but do not execute them' )
group_test.add_argument( '--start',   action='store', help='routine to start with, helpful for restarting', default='' )
group_test.add_argument( '-x', '--exclude', action='append', help='routines to exclude; repeatable', default=[] )

group_size = parser.add_argument_group( 'matrix dimensions (default is medium)' )
group_size.add_argument( '--quick',  action='store_true', help='run quick "sanity check" of few, small tests' )
group_size.add_argument( '--xsmall', action='store_true', help='run x-small tests' )
group_size.add_argument( '--small',  action='store_true', help='run small tests' )
group_size.add_argument( '--medium', action='store_true', help='run medium tests' )
group_size.add_argument( '--large',  action='store_true', help='run large tests' )
group_size.add_argument( '--square', action='store_true', help='run square (m = n = k) tests', default=False )
group_size.add_argument( '--tall',   action='store_true', help='run tall (m > n) tests', default=False )
group_size.add_argument( '--wide',   action='store_true', help='run wide (m < n) tests', default=False )
group_size.add_argument( '--mnk',    action='store_true', help='run tests with m, n, k all different', default=False )
group_size.add_argument( '--dim',    action='store',      help='explicitly specify size', default='' )

group_cat = parser.add_argument_group( 'category (default is all)' )
categories = [
    group_cat.add_argument( '--lu',            action='store_true', help='run LU tests' ),
    group_cat.add_argument( '--gb',            action='store_true', help='run GB tests' ),
    group_cat.add_argument( '--gt',            action='store_true', help='run GT tests' ),
    group_cat.add_argument( '--chol',          action='store_true', help='run Cholesky tests' ),
    group_cat.add_argument( '--sysv',          action='store_true', help='run symmetric indefinite (Bunch-Kaufman) tests' ),
    group_cat.add_argument( '--rook',          action='store_true', help='run symmetric indefinite (rook) tests' ),
    group_cat.add_argument( '--aasen',         action='store_true', help='run symmetric indefinite (Aasen) tests' ),
    group_cat.add_argument( '--hesv',          action='store_true', help='run hermetian tests (FIXME more informationhere)' ),
    group_cat.add_argument( '--least-squares', action='store_true', help='run least squares tests' ),
    group_cat.add_argument( '--qr',            action='store_true', help='run QR tests' ),
    group_cat.add_argument( '--lq',            action='store_true', help='run LQ tests' ),
    group_cat.add_argument( '--ql',            action='store_true', help='run QL tests' ),
    group_cat.add_argument( '--rq',            action='store_true', help='run RQ tests' ),
    group_cat.add_argument( '--syev',          action='store_true', help='run symmetric eigenvalues tests' ),
    group_cat.add_argument( '--sygv',          action='store_true', help='run generalized symmetric eigenvalues tests' ),
    group_cat.add_argument( '--geev',          action='store_true', help='run non-symmetric eigenvalues tests' ),
    group_cat.add_argument( '--svd',           action='store_true', help='run svd tests' ),
    group_cat.add_argument( '--aux',           action='store_true', help='run auxiliary tests' ),
    group_cat.add_argument( '--aux-house',     action='store_true', help='run auxiliary Householder reflector tests' ),
    group_cat.add_argument( '--aux-givens',    action='store_true', help='run auxiliary Givens rotations tests' ),
    group_cat.add_argument( '--aux-norm',      action='store_true', help='run auxiliary norm tests' ),
    group_cat.add_argument( '--blas',          action='store_true', help='run additional BLAS tests' ),
]
# map category objects to category names: ['lu', 'chol', ...]
categories = list( map( lambda x: x.dest, categories ) )

group_target = parser.add_argument_group( 'target' )
group_target.add_argument( '--host',          action='store_true', help='run all CPU host routines' ),
group_target.add_argument( '--device',        action='store_true', help='run all GPU device routines' ),

group_opt = parser.add_argument_group( 'options' )
# BLAS and LAPACK
# Empty defaults (check, ref, etc.) use the default in test.cc.
group_opt.add_argument( '--type',   action='store', help='default=%(default)s', default='s,d,c,z' )
group_opt.add_argument( '--layout', action='store', help='default=%(default)s', default='c,r' )
group_opt.add_argument( '--transA', action='store', help='default=%(default)s', default='n,t,c' )
group_opt.add_argument( '--transB', action='store', help='default=%(default)s', default='n,t,c' )
group_opt.add_argument( '--trans',  action='store', help='default=%(default)s', default='n,t,c' )
group_opt.add_argument( '--uplo',   action='store', help='default=%(default)s', default='l,u' )
group_opt.add_argument( '--diag',   action='store', help='default=%(default)s', default='n,u' )
group_opt.add_argument( '--side',   action='store', help='default=%(default)s', default='l,r' )
group_opt.add_argument( '--alpha',  action='store', help='default=%(default)s', default='' )
group_opt.add_argument( '--beta',   action='store', help='default=%(default)s', default='' )
group_opt.add_argument( '--incx',   action='store', help='default=%(default)s', default='1,2,-1,-2' )
group_opt.add_argument( '--incy',   action='store', help='default=%(default)s', default='1,2,-1,-2' )
group_opt.add_argument( '--align',  action='store', help='default=%(default)s', default='32' )
group_opt.add_argument( '--check',  action='store', help='default=y', default='' )  # default in test.cc
group_opt.add_argument( '--ref',    action='store', help='default=y', default='' )  # default in test.cc
group_opt.add_argument( '--verbose', action='store', help='default=0', default='' )  # default in test.cc

# LAPACK only
group_opt.add_argument( '--itype',  action='store', help='default=%(default)s', default='1,2,3' )
group_opt.add_argument( '--factored', action='store', help='default=%(default)s', default='f,n,e' )
group_opt.add_argument( '--equed',  action='store', help='default=%(default)s', default='n,r,c,b' )
group_opt.add_argument( '--pivot',  action='store', help='default=%(default)s', default='v,t,b' )
group_opt.add_argument( '--direction', action='store', help='default=%(default)s', default='f,b' )
group_opt.add_argument( '--storev', action='store', help='default=%(default)s', default='c,r' )
group_opt.add_argument( '--norm',   action='store', help='default=%(default)s', default='max,1,inf,fro' )
group_opt.add_argument( '--ijob',   action='store', help='default=%(default)s', default='0:5:1' )
group_opt.add_argument( '--jobz',   action='store', help='default=%(default)s', default='n,v' )
group_opt.add_argument( '--jobvl',  action='store', help='default=%(default)s', default='n,v' )
group_opt.add_argument( '--jobvr',  action='store', help='default=%(default)s', default='n,v' )
group_opt.add_argument( '--jobvs',  action='store', help='default=%(default)s', default='n,v' )
group_opt.add_argument( '--jobu',   action='store', help='default=%(default)s', default='n,s,o,a' )
group_opt.add_argument( '--jobvt',  action='store', help='default=%(default)s', default='n,s,o,a' )
group_opt.add_argument( '--balanc', action='store', help='default=%(default)s', default='n,p,s,b' )
group_opt.add_argument( '--sort',   action='store', help='default=%(default)s', default='n,s' )
group_opt.add_argument( '--select', action='store', help='default=%(default)s', default='n,s' )
group_opt.add_argument( '--sense',  action='store', help='default=%(default)s', default='n,e,v,b' )
group_opt.add_argument( '--vect',   action='store', help='default=%(default)s', default='n,v' )
group_opt.add_argument( '--l',      action='store', help='default=%(default)s', default='0,100' )
group_opt.add_argument( '--ka',     action='store', help='default=%(default)s', default='20,100' )
group_opt.add_argument( '--kb',     action='store', help='default=%(default)s', default='20,100' )
group_opt.add_argument( '--kd',     action='store', help='default=%(default)s', default='20,100' )
group_opt.add_argument( '--kl',     action='store', help='default=%(default)s', default='20,100' )
group_opt.add_argument( '--ku',     action='store', help='default=%(default)s', default='20,100' )
group_opt.add_argument( '--vl',     action='store', help='default=%(default)s', default='-inf,0' )
group_opt.add_argument( '--vu',     action='store', help='default=%(default)s', default='inf' )
group_opt.add_argument( '--il',     action='store', help='default=%(default)s', default='10' )
group_opt.add_argument( '--iu',     action='store', help='default=%(default)s', default='-1,100' )
group_opt.add_argument( '--nb',     action='store', help='default=%(default)s', default='64' )
group_opt.add_argument( '--matrixtype', action='store', help='default=%(default)s', default='g,l,u' )

parser.add_argument( 'tests', nargs=argparse.REMAINDER )
opts = parser.parse_args()

for t in opts.tests:
    if (t.startswith('--')):
        print( 'Error: option', t, 'must come before any routine names' )
        print( 'usage:', sys.argv[0], '[options]', '[routines]' )
        print( '      ', sys.argv[0], '--help' )
        exit(1)

# by default, run medium sizes
if (not (opts.quick or opts.xsmall or opts.small or opts.medium or opts.large)):
    opts.medium = True

# by default, run all shapes
if (not (opts.square or opts.tall or opts.wide or opts.mnk)):
    opts.square = True
    opts.tall   = True
    opts.wide   = True
    opts.mnk    = True

# By default, run both host and device.
if (not opts.host and not opts.device):
    opts.host   = True
    opts.device = True

# By default, or if specific test routines given, enable all categories
# to get whichever has the routines.
if (opts.tests or not any( map( lambda c: opts.__dict__[ c ], categories ))):
    for c in categories:
        opts.__dict__[ c ] = True

start_routine = opts.start

# ------------------------------------------------------------------------------
# parameters
# begin with space to ease concatenation

# if given, use explicit dim
dim = ' --dim ' + opts.dim if (opts.dim) else ''
n        = dim
tall     = dim
wide     = dim
mn       = dim
mnk      = dim
nk_tall  = dim
nk_wide  = dim
nk       = dim

if (not opts.dim):
    if (opts.quick):
        n        = ' --dim 100'
        tall     = ' --dim 100x50'  # 2:1
        wide     = ' --dim 50x100'  # 1:2
        mnk      = ' --dim 25x50x75'
        nk_tall  = ' --dim 1x100x50'  # 2:1
        nk_wide  = ' --dim 1x50x100'  # 1:2
        opts.incx  = '1,-1'
        opts.incy  = '1,-1'
        opts.batch = '10'
        opts.l     = '0,20,50'
        opts.nb    = '16'

    if (opts.xsmall):
        n       += ' --dim 10'
        tall    += ' --dim 20x10'
        wide    += ' --dim 10x20'
        mnk     += ' --dim 10x15x20 --dim 15x10x20' \
                +  ' --dim 10x20x15 --dim 15x20x10' \
                +  ' --dim 20x10x15 --dim 20x15x10'
        nk_tall += ' --dim 1x20x10'
        nk_wide += ' --dim 1x10x20'
        # tpqrt, tplqt needs small l, nb <= min( m, n )
        if (opts.l == parser.get_default('l')):
            opts.l = '0,5,100'
        if (opts.nb == parser.get_default('nb')):
            opts.nb = '8,64'
        if (opts.ka == parser.get_default('ka')):
            opts.ka = '5'
        if (opts.kb == parser.get_default('kb')):
            opts.kb = '5'
        if (opts.kd == parser.get_default('kd')):
            opts.kd = '5'

    if (opts.small):
        n       += ' --dim 25:100:25'
        tall    += ' --dim 50:200:50x25:100:25'  # 2:1
        wide    += ' --dim 25:100:25x50:200:50'  # 1:2
        mnk     += ' --dim 25x50x75 --dim 50x25x75' \
                +  ' --dim 25x75x50 --dim 50x75x25' \
                +  ' --dim 75x25x50 --dim 75x50x25'
        nk_tall += ' --dim 1x50:200:50x25:100:25'
        nk_wide += ' --dim 1x25:100:25x50:200:50'

    if (opts.medium):
        n       += ' --dim 100:500:100'
        tall    += ' --dim 200:1000:200x100:500:100'  # 2:1
        wide    += ' --dim 100:500:100x200:1000:200'  # 1:2
        mnk     += ' --dim 100x300x600 --dim 300x100x600' \
                +  ' --dim 100x600x300 --dim 300x600x100' \
                +  ' --dim 600x100x300 --dim 600x300x100'
        nk_tall += ' --dim 1x200:1000:200x100:500:100'
        nk_wide += ' --dim 1x100:500:100x200:1000:200'

    if (opts.large):
        n       += ' --dim 1000:5000:1000'
        tall    += ' --dim 2000:10000:2000x1000:5000:1000'  # 2:1
        wide    += ' --dim 1000:5000:1000x2000:10000:2000'  # 1:2
        mnk     += ' --dim 1000x3000x6000 --dim 3000x1000x6000' \
                +  ' --dim 1000x6000x3000 --dim 3000x6000x1000' \
                +  ' --dim 6000x1000x3000 --dim 6000x3000x1000'
        nk_tall += ' --dim 1x2000:10000:2000x1000:5000:1000'
        nk_wide += ' --dim 1x1000:5000:1000x2000:10000:2000'

    mn  = ''
    nk  = ''
    if (opts.square):
        mn = n
        nk = n
    if (opts.tall):
        mn += tall
        nk += nk_tall
    if (opts.wide):
        mn += wide
        nk += nk_wide
    if (opts.mnk):
        mnk = mn + mnk
    else:
        mnk = mn
# end

# BLAS and LAPACK
dtype  = ' --type '   + opts.type   if (opts.type)   else ''
layout = ' --layout ' + opts.layout if (opts.layout) else ''
transA = ' --transA ' + opts.transA if (opts.transA) else ''
transB = ' --transB ' + opts.transB if (opts.transB) else ''
trans  = ' --trans '  + opts.trans  if (opts.trans)  else ''
uplo   = ' --uplo '   + opts.uplo   if (opts.uplo)   else ''
diag   = ' --diag '   + opts.diag   if (opts.diag)   else ''
side   = ' --side '   + opts.side   if (opts.side)   else ''
a      = ' --alpha '  + opts.alpha  if (opts.alpha)  else ''
ab     = a+' --beta ' + opts.beta   if (opts.beta)   else a
incx   = ' --incx '   + opts.incx   if (opts.incx)   else ''
incy   = ' --incy '   + opts.incy   if (opts.incy)   else ''
align  = ' --align '  + opts.align  if (opts.align)  else ''
check  = ' --check '  + opts.check  if (opts.check)  else ''
ref    = ' --ref '    + opts.ref    if (opts.ref)    else ''
verbose = ' --verbose ' + opts.verbose if (opts.verbose) else ''

# LAPACK only
itype  = ' --itype '  + opts.itype  if (opts.itype)  else ''
factored = ' --factored ' + opts.factored if (opts.factored)  else ''
equed  = ' --equed '  + opts.equed  if (opts.equed)  else ''
pivot  = ' --pivot '  + opts.pivot  if (opts.pivot)  else ''
direction = ' --direction ' + opts.direction if (opts.direction) else ''
storev = ' --storev ' + opts.storev if (opts.storev) else ''
norm   = ' --norm '   + opts.norm   if (opts.norm)   else ''
ijob   = ' --ijob '   + opts.ijob   if (opts.ijob)   else ''
jobz   = ' --jobz '   + opts.jobz   if (opts.jobz)   else ''
jobu   = ' --jobu '   + opts.jobu   if (opts.jobu)   else ''
jobvt  = ' --jobvt '  + opts.jobvt  if (opts.jobvt)  else ''
jobvl  = ' --jobvl '  + opts.jobvl  if (opts.jobvl)  else ''
jobvr  = ' --jobvr '  + opts.jobvr  if (opts.jobvr)  else ''
jobvs  = ' --jobvs '  + opts.jobvs  if (opts.jobvs)  else ''
balanc = ' --balanc ' + opts.balanc if (opts.balanc) else ''
sort   = ' --sort '   + opts.sort   if (opts.sort)   else ''
sense  = ' --sense '  + opts.sense  if (opts.sense)  else ''
vect   = ' --vect '   + opts.vect   if (opts.vect)   else ''
l      = ' --l '      + opts.l      if (opts.l)      else ''
nb     = ' --nb '     + opts.nb     if (opts.nb)     else ''
ka     = ' --ka '     + opts.ka     if (opts.ka)     else ''
kb     = ' --kb '     + opts.kb     if (opts.kb)     else ''
kd     = ' --kd '     + opts.kd     if (opts.kd)     else ''
kl     = ' --kl '     + opts.kl     if (opts.kl)     else ''
ku     = ' --ku '     + opts.ku     if (opts.ku)     else ''
vl     = ' --vl '     + opts.vl     if (opts.vl)     else ''
vu     = ' --vu '     + opts.vu     if (opts.vu)     else ''
il     = ' --il '     + opts.il     if (opts.il)     else ''
iu     = ' --iu '     + opts.iu     if (opts.iu)     else ''
mtype  = ' --matrixtype ' + opts.matrixtype if (opts.matrixtype) else ''

# general options for all routines
gen = check + ref + verbose

# ------------------------------------------------------------------------------
# filters a comma separated list csv based on items in list values.
# if no items from csv are in values, returns first item in values.
def filter_csv( values, csv ):
    f = list( filter( lambda x: x in values, csv.split( ',' ) ) )
    if (not f):
        return values[0]
    return ','.join( f )
# end

# ------------------------------------------------------------------------------
# limit options to specific values
dtype_real    = ' --type ' + filter_csv( ('s', 'd'), opts.type )
dtype_complex = ' --type ' + filter_csv( ('c', 'z'), opts.type )
dtype_double  = ' --type ' + filter_csv( ('d', 'z'), opts.type )

trans_nt = ' --trans ' + filter_csv( ('n', 't'), opts.trans )
trans_nc = ' --trans ' + filter_csv( ('n', 'c'), opts.trans )

# positive inc
incx_pos = ' --incx ' + filter_csv( ('1', '2'), opts.incx )
incy_pos = ' --incy ' + filter_csv( ('1', '2'), opts.incy )

# ------------------------------------------------------------------------------
cmds = []

# LU
if (opts.lu and opts.host):
    cmds += [
    [ 'gesv',  gen + dtype + align + n ],
    # todo: equed
    [ 'gesvx', gen + dtype + align + n + factored + trans ],
    [ 'getrf', gen + dtype + align + mn ],
    [ 'getrs', gen + dtype + align + n + trans ],
    [ 'getri', gen + dtype + align + n ],
    [ 'gecon', gen + dtype + align + n ],
    [ 'gerfs', gen + dtype + align + n + trans ],
    [ 'geequ', gen + dtype + align + n ],
    ]

if (opts.lu and opts.device):
    # GPU
    cmds += [
    [ 'dev-getrf', gen + dtype + align + n ],
    ]

# General Banded
if (opts.gb and opts.host):
    cmds += [
    [ 'gbsv',  gen + dtype + align + n  + kl + ku ],
    [ 'gbtrf', gen + dtype + align + mn + kl + ku ],
    [ 'gbtrs', gen + dtype + align + n  + kl + ku + trans ],
    [ 'gbcon', gen + dtype + align + n  + kl + ku ],
    [ 'gbrfs', gen + dtype + align + n  + kl + ku + trans ],
    [ 'gbequ', gen + dtype + align + n  + kl + ku ],
    ]

# General Tri-Diagonal
if (opts.gt and opts.host):
    cmds += [
    [ 'gtsv',  gen + dtype + align + n ],
    [ 'gttrf', gen + dtype +         n ],
    [ 'gttrs', gen + dtype + align + n + trans ],
    [ 'gtcon', gen + dtype +         n ],
    [ 'gtrfs', gen + dtype + align + n + trans ],
    ]

# Cholesky
if (opts.chol and opts.host):
    cmds += [
    [ 'posv',  gen + dtype + align + n + uplo ],
    [ 'potrf', gen + dtype + align + n + uplo ],
    [ 'potrs', gen + dtype + align + n + uplo ],
    [ 'potri', gen + dtype + align + n + uplo ],
    [ 'pocon', gen + dtype + align + n + uplo ],
    [ 'porfs', gen + dtype + align + n + uplo ],
    [ 'poequ', gen + dtype + align + n ],  # only diagonal elements (no uplo)

    # Packed
    [ 'ppsv',  gen + dtype + align + n + uplo ],
    [ 'pptrf', gen + dtype +         n + uplo ],
    [ 'pptrs', gen + dtype + align + n + uplo ],
    [ 'pptri', gen + dtype +         n + uplo ],
    [ 'ppcon', gen + dtype +         n + uplo ],
    [ 'pprfs', gen + dtype + align + n + uplo ],
    [ 'ppequ', gen + dtype +         n + uplo ],

    # Banded
    [ 'pbsv',  gen + dtype + align + n + kd + uplo ],
    [ 'pbtrf', gen + dtype + align + n + kd + uplo ],
    [ 'pbtrs', gen + dtype + align + n + kd + uplo ],
    [ 'pbcon', gen + dtype + align + n + kd + uplo ],
    [ 'pbrfs', gen + dtype + align + n + kd + uplo ],
    [ 'pbequ', gen + dtype + align + n + kd + uplo ],

    # Tri-diagonal
    [ 'ptsv',  gen + dtype + align + n ],
    [ 'pttrf', gen + dtype         + n ],
    [ 'pttrs', gen + dtype + align + n + uplo ],
    [ 'ptcon', gen + dtype         + n ],
    [ 'ptrfs', gen + dtype + align + n + uplo ],
    ]

if (opts.chol and opts.device):
    # GPU
    cmds += [
    [ 'dev-potrf', gen + dtype + align + n + uplo ],
    ]

# symmetric indefinite, Bunch-Kaufman
if (opts.sysv and opts.host):
    cmds += [
    [ 'sysv',  gen + dtype + align + n + uplo ],
    [ 'sytrf', gen + dtype + align + n + uplo ],
    [ 'sytrs', gen + dtype + align + n + uplo ],
    [ 'sytri', gen + dtype + align + n + uplo ],
    [ 'sycon', gen + dtype + align + n + uplo ],
    [ 'syrfs', gen + dtype + align + n + uplo ],

    # Packed
    [ 'spsv',  gen + dtype + align + n + uplo ],
    [ 'sptrf', gen + dtype         + n + uplo ],
    [ 'sptrs', gen + dtype + align + n + uplo ],
    [ 'sptri', gen + dtype         + n + uplo ],
    [ 'spcon', gen + dtype         + n + uplo ],
    [ 'sprfs', gen + dtype + align + n + uplo ],
    ]

# symmetric indefinite, rook
if (opts.rook and opts.host):
    cmds += [
    # original Rook
    [ 'sysv_rook',  gen + dtype + align + n + uplo ],
    [ 'sytrf_rook', gen + dtype + align + n + uplo ],
    [ 'sytrs_rook', gen + dtype + align + n + uplo ],
    #[ 'sytri_rook', gen + dtype + align + n + uplo ],

    # new Rook
    [ 'sysv_rk',    gen + dtype + align + n + uplo ],
    [ 'sytrf_rk',   gen + dtype + align + n + uplo ],
    #[ 'sytrs_rk',   gen + dtype + align + n + uplo ],
    #[ 'sytri_rk',   gen + dtype + align + n + uplo ],
    ]

# symmetric indefinite, Aasen
if (opts.aasen and opts.host):
    cmds += [
    [ 'sysv_aa',  gen + dtype + align + n + uplo ],
    [ 'sytrf_aa', gen + dtype + align + n + uplo ],
    [ 'sytrs_aa', gen + dtype + align + n + uplo ],
    #[ 'sytri_aa', gen + dtype + align + n + uplo ],

    #[ 'sysv_aasen_2stage',  gen + dtype + align + n + uplo ],
    #[ 'sytrf_aasen_2stage', gen + dtype + align + n + uplo ],
    #[ 'sytrs_aasen_2stage', gen + dtype + align + n + uplo ],
    #[ 'sytri_aasen_2stage', gen + dtype + align + n + uplo ],
    ]

# Hermitian indefinite
if (opts.hesv and opts.host):
    cmds += [
    [ 'hesv',  gen + dtype + align + n + uplo ],
    [ 'hetrf', gen + dtype + align + n + uplo ],
    [ 'hetrs', gen + dtype + align + n + uplo ],
    [ 'hetri', gen + dtype + align + n + uplo ],
    [ 'hecon', gen + dtype + align + n + uplo ],
    [ 'herfs', gen + dtype + align + n + uplo ],

    # Packed
    [ 'hpsv',  gen + dtype + align + n + uplo ],
    [ 'hptrf', gen + dtype + n + uplo ],
    [ 'hptrs', gen + dtype + align + n + uplo ],
    [ 'hptri', gen + dtype + n + uplo ],
    [ 'hpcon', gen + dtype + n + uplo ],
    [ 'hprfs', gen + dtype + align + n + uplo ],
    ]

# least squares
if (opts.least_squares and opts.host):
    cmds += [
    [ 'gels',   gen + dtype + align + mn + trans_nc ],
    [ 'gelsy',  gen + dtype + align + mn ],
    # todo: gelsd is failing
    #[ 'gelsd',  gen + dtype + align + mn ],
    [ 'gelss',  gen + dtype + align + mn ],
    [ 'getsls', gen + dtype + align + mn + trans_nc ],

    # Generalized
    [ 'gglse', gen + dtype + align + mnk ],
    # todo: ggglm is failing
    #[ 'ggglm', gen + dtype + align + mnk ],
    ]

# QR
if (opts.qr and opts.host):
    cmds += [
    [ 'geqr',  gen + dtype + align + n + wide + tall ],
    [ 'geqrf', gen + dtype + align + n + wide + tall ],
    # todo: ggqrf is failing
    #[ 'ggqrf', gen + dtype + align + mnk ],
    [ 'ungqr', gen + dtype + align + mn ],  # m >= n
    #[ 'unmqr', gen + dtype_real    + align + mnk + side + trans    ],  # real does trans = N, T, C
    #[ 'unmqr', gen + dtype_complex + align + mnk + side + trans_nc ],  # complex does trans = N, C, not T

    [ 'orhr_col', gen + dtype_real + align + n + tall ],
    [ 'unhr_col', gen + dtype      + align + n + tall ],

    [ 'gemqrt', gen + dtype_real    + align + n + nb + side + trans    ],  # real does trans = N, T, C
    [ 'gemqrt', gen + dtype_complex + align + n + nb + side + trans_nc ],  # complex does trans = N, C, not T

    # Triangle-pentagon
    [ 'tpqrt',  gen + dtype + align + mn + l + nb ],
    [ 'tpqrt2', gen + dtype + align + mn + l ],
    [ 'tpmqrt', gen + dtype_real    + align + mn + l + nb + side + trans    ],  # real does trans = N, T, C
    [ 'tpmqrt', gen + dtype_complex + align + mn + l + nb + side + trans_nc ],  # complex does trans = N, C, not T
    #[ 'tprfb',  gen + dtype + align + mn + l ],  # TODO: bug in LAPACKE crashes tester
    ]

if (opts.qr and opts.device):
    # GPU
    cmds += [
    [ 'dev-geqrf', gen + dtype + align + n + wide + tall ],
    ]

# LQ
if (opts.lq and opts.host):
    cmds += [
    [ 'gelqf', gen + dtype + align + mn ],
    #[ 'gglqf', gen + dtype + align + mn ],
    [ 'unglq', gen + dtype + align + mn ],  # m <= n, k <= m  TODO Fix the input sizes to match constraints
    #[ 'unmlq', gen + dtype_real    + align + mnk + side + trans    ],  # real does trans = N, T, C
    #[ 'unmlq', gen + dtype_complex + align + mnk + side + trans_nc ],  # complex does trans = N, C, not T

    # Triangle-pentagon
    [ 'tplqt',  gen + dtype + align + mn + l + nb ],
    [ 'tplqt2', gen + dtype + align + mn + l ],
    [ 'tpmlqt', gen + dtype_real    + align + mn + l + nb + side + trans    ],  # real does trans = N, T, C
    [ 'tpmlqt', gen + dtype_complex + align + mn + l + nb + side + trans_nc ],  # complex does trans = N, C, not T
    ]

# QL
if (opts.ql and opts.host):
    cmds += [
    [ 'geqlf', gen + dtype + align + mn ],
    #[ 'ggqlf', gen + dtype + align + mn ],
    [ 'ungql', gen + dtype + align + mn ],
    #[ 'unmql', gen + dtype_real    + align + mnk + side + trans    ],  # real does trans = N, T, C
    #[ 'unmql', gen + dtype_complex + align + mnk + side + trans_nc ],  # complex does trans = N, C, not T
    ]

# RQ
if (opts.rq and opts.host):
    cmds += [
    [ 'gerqf', gen + dtype + align + mn ],
    # todo: ggrqf is failing
    #[ 'ggrqf', gen + dtype + align + mnk ],
    [ 'ungrq', gen + dtype + align + mnk ],
    #[ 'unmrq', gen + dtype_real    + align + mnk + side + trans    ],  # real does trans = N, T, C
    #[ 'unmrq', gen + dtype_complex + align + mnk + side + trans_nc ],  # complex does trans = N, C, not T
    ]

# symmetric eigenvalues
if (opts.syev and opts.host):
    cmds += [
    [ 'heev',  gen + dtype + align + n + jobz + uplo ],
    [ 'heevx', gen + dtype + align + n + jobz + uplo + vl + vu ],
    [ 'heevx', gen + dtype + align + n + jobz + uplo + il + iu ],
    [ 'heevd', gen + dtype + align + n + jobz + uplo ],
    [ 'heevr', gen + dtype + align + n + jobz + uplo + vl + vu ],
    [ 'heevr', gen + dtype + align + n + jobz + uplo + il + iu ],
    [ 'hetrd', gen + dtype + align + n + uplo ],
    [ 'lae2',  gen + dtype_real ],  # 2x2, eigvals only
    [ 'laev2', gen + dtype ],  # 2x2
    [ 'ungtr', gen + dtype + align + n + uplo ],
    [ 'unmtr', gen + dtype_real    + align + mn + uplo + side + trans    ],  # real does trans = N, T, C
    [ 'unmtr', gen + dtype_complex + align + mn + uplo + side + trans_nc ],  # complex does trans = N, C, not T

    # Packed
    [ 'hpev',  gen + dtype + align + n + jobz + uplo ],
    [ 'hpevx', gen + dtype + align + n + jobz + uplo + vl + vu ],
    [ 'hpevx', gen + dtype + align + n + jobz + uplo + il + iu ],
    [ 'hpevd', gen + dtype + align + n + jobz + uplo ],
    #[ 'hpevr', gen + dtype + align + n + jobz + uplo + vl + vu ],
    #[ 'hpevr', gen + dtype + align + n + jobz + uplo + il + iu ],
    [ 'hptrd', gen + dtype + n + uplo ],
    [ 'upgtr', gen + dtype + align + n + uplo ],
    [ 'upmtr', gen + dtype + align + mn + side + uplo + trans_nc ],

    # Banded
    [ 'hbev',  gen + dtype + align + n + jobz + uplo ],
    [ 'hbevx', gen + dtype + align + n + jobz + uplo + vl + vu ],
    [ 'hbevx', gen + dtype + align + n + jobz + uplo + il + iu ],
    [ 'hbevd', gen + dtype + align + n + jobz + uplo ],
    #[ 'hbevr', gen + dtype + align + n + jobz + uplo + vl + vu ],
    #[ 'hbevr', gen + dtype + align + n + jobz + uplo + il + iu ],
    #[ 'hbtrd', gen + dtype + align + n + uplo ],
    #[ 'ubgtr', gen + dtype + align + n + uplo ],
    #[ 'ubmtr', gen + dtype_real    + la + mn + uplo + side + trans    ],
    #[ 'ubmtr', gen + dtype_complex + la + mn + uplo + side + trans_nc ],
    ]

if (opts.syev and opts.device):
    # GPU
    cmds += [
    [ 'dev-heevd', gen + dtype + align + n + jobz + uplo ],
    ]

# generalized symmetric eigenvalues
if (opts.sygv and opts.host):
    cmds += [
    [ 'hegv',  gen + dtype + align + n + itype + jobz + uplo ],
    [ 'hegvx', gen + dtype + align + n + itype + jobz + uplo + vl + vu ],
    [ 'hegvx', gen + dtype + align + n + itype + jobz + uplo + il + iu ],
    [ 'hegvd', gen + dtype + align + n + itype + jobz + uplo ],
    #[ 'hegvr', gen + dtype + align + n + uplo ],
    [ 'hegst', gen + dtype + align + n + itype + uplo ],

    # Packed
    [ 'hpgv',  gen + dtype + align + n + itype + jobz + uplo ],
    [ 'hpgvx', gen + dtype + align + n + itype + jobz + uplo + vl + vu ],
    [ 'hpgvx', gen + dtype + align + n + itype + jobz + uplo + il + iu ],
    [ 'hpgvd', gen + dtype + align + n + itype + jobz + uplo ],
    #[ 'hpgvr', gen + dtype + align + n + uplo ],
    [ 'hpgst', gen + dtype + n + itype + uplo ],

    # Banded
    [ 'hbgv',  gen + dtype + align + n + jobz + uplo + ka + kb ],
    [ 'hbgvx', gen + dtype + align + n + jobz + uplo + ka + kb + vl + vu ],
    [ 'hbgvx', gen + dtype + align + n + jobz + uplo + ka + kb + il + iu ],
    [ 'hbgvd', gen + dtype + align + n + jobz + uplo + ka + kb ],
    #[ 'hbgvr', gen + dtype + align + n + uplo + ka + kb ],
    #[ 'hbgst', gen + dtype + align + n + vect + uplo + ka + kb ],
    ]

# non-symmetric eigenvalues
if (opts.geev and opts.host):
    cmds += [
    [ 'geev',  gen + dtype + align + n + jobvl + jobvr ],
    # todo: ggev is failing
    #[ 'ggev',  gen + dtype + align + n + jobvl + jobvr ],
    #[ 'geevx', gen + dtype + align + n + balanc + jobvl + jobvr + sense ],
    [ 'gehrd', gen + dtype + align + n ],
    [ 'unghr', gen + dtype + align + n ],
    [ 'unmhr', gen + dtype_real    + align + mn + side + trans    ],  # real does trans = N, T, C
    [ 'unmhr', gen + dtype_complex + align + mn + side + trans_nc ],  # complex does trans = N, C, not T
    #[ 'trevc', gen + dtype + align + n + side + howmany + select ],
    #[ 'geesx', gen + dtype + align + n + jobvs + sort + select + sense ],
    [ 'tgexc', gen + dtype + align + n + jobvl + jobvr ],
    [ 'tgsen', gen + dtype + align + n + jobvl + jobvr + ijob ],
    ]

# svd
if (opts.svd and opts.host):
    cmds += [
    # todo: MKL seems to have a bug with jobu=o,s and jobvt=o,s,a
    # for tall matrices, e.g., dim=100x50. Skip failing combinations for now.
    #[ 'gesvd',         gen + dtype + align + mn + jobu + jobvt ],
    [ 'gesvd',         gen + dtype + align + mn + " --jobu n,a" + jobvt ],
    [ 'gesvd',         gen + dtype + align + mn + " --jobu o,s --jobvt n" ],
    [ 'gesdd',         gen + dtype + align + mn + jobu ],
    # todo: gesvdx is failing
    #[ 'gesvdx',        gen + dtype + align + mn + jobz + jobvr + vl + vu ],
    #[ 'gesvdx',        gen + dtype + align + mn + jobz + jobvr + il + iu ],
    #[ 'gesvd_2stage',  gen + dtype + align + mn ],
    #[ 'gesdd_2stage',  gen + dtype + align + mn ],
    #[ 'gesvdx_2stage', gen + dtype + align + mn ],
    #[ 'gejsv',         gen + dtype + align + mn ],
    #[ 'gesvj',         gen + dtype + align + mn + joba + jobu + jobv ],
    ]

# auxilary
if (opts.aux and opts.host):
    cmds += [
    [ 'lacpy', gen + dtype + align + mn + mtype ],
    [ 'laed4', gen + dtype_real + n ],
    [ 'laset', gen + dtype + align + mn + mtype ],
    [ 'laswp', gen + dtype + align + mn ],
    ]

# auxilary - householder
if (opts.aux_house and opts.host):
    cmds += [
    [ 'larfg', dtype         + n   + incx_pos ],
    [ 'larfgp', dtype        + n   + incx_pos ],
    [ 'larf',  gen + dtype + align + mn  + incx + side ],
    [ 'larfx', gen + dtype + align + mn  + side ],
    [ 'larfy', gen + dtype + align + n   + incx ],
    [ 'larfb', gen + dtype + align + mnk + side + trans + direction + storev ],
    [ 'larft', gen + dtype + align + nk  + direction + storev ],
    ]

# auxilary - Givens rotations
if (opts.aux_givens and opts.host):
    cmds += [
    [ 'lasr', gen + dtype + align + mn + side + pivot + direction ],
    ]

# auxilary - norms
if (opts.aux_norm and opts.host):
    cmds += [
    [ 'lange', gen + dtype + align + mn + norm ],
    # todo: lanhe is failing
    #[ 'lanhe', gen + dtype + align + n  + norm + uplo ],
    [ 'lansy', gen + dtype + align + n  + norm + uplo ],
    [ 'lantr', gen + dtype + align + mn + norm + uplo + diag ],
    [ 'lanhs', gen + dtype + align + n  + norm ],

    # Packed
    [ 'lanhp', gen + dtype + n + norm + uplo ],
    [ 'lansp', gen + dtype + n + norm + uplo ],
    [ 'lantp', gen + dtype + n + norm + uplo + diag ],

    # Banded
    [ 'langb', gen + dtype + align + mn + kl + ku + norm ],
    [ 'lanhb', gen + dtype + align + n + kd + norm + uplo ],
    [ 'lansb', gen + dtype + align + n + kd + norm + uplo ],
    [ 'lantb', gen + dtype + align + n + kd + norm + uplo + diag ],

    # Tri-diagonal
    [ 'langt', gen + dtype + n + norm ],
    [ 'lanht', gen + dtype + n + norm ],
    [ 'lanst', gen + dtype + n + norm ],
    ]

# additional blas
if (opts.blas and opts.host):
    cmds += [
    [ 'syr',   gen + dtype + align + n + uplo ],
    [ 'symv',  gen + dtype + layout + align + uplo + n + incx + incy ],
    ]

# ------------------------------------------------------------------------------
# When stdout is redirected to file instead of TTY console,
# and  stderr is still going to a TTY console,
# print extra summary messages to stderr.
output_redirected = sys.stderr.isatty() and not sys.stdout.isatty()

# ------------------------------------------------------------------------------
# if output is redirected, prints to both stderr and stdout;
# otherwise prints to just stdout.
def print_tee( *args ):
    global output_redirected
    print( *args )
    if (output_redirected):
        print( *args, file=sys.stderr )
# end

# ------------------------------------------------------------------------------
# cmd is a pair of strings: (function, args)

def run_test( cmd ):
    cmd = opts.test +' '+ cmd[1] +' '+ cmd[0]
    print_tee( cmd )
    if (opts.dry_run):
        return (None, None)

    output = ''
    p = subprocess.Popen( cmd.split(), stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT )
    p_out = p.stdout
    if (sys.version_info.major >= 3):
        p_out = io.TextIOWrapper(p.stdout, encoding='utf-8')
    # Read unbuffered ("for line in p.stdout" will buffer).
    for line in iter(p_out.readline, ''):
        print( line, end='' )
        output += line
    err = p.wait()
    if (err != 0):
        print_tee( 'FAILED: exit code', err )
    else:
        print_tee( 'pass' )
    return (err, output)
# end

# ------------------------------------------------------------------------------
# Utility to pretty print XML.
# See https://stackoverflow.com/a/33956544/1655607
#
def indent_xml( elem, level=0 ):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent_xml( elem, level+1 )
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i
# end

# ------------------------------------------------------------------------------
# run each test

start = time.time()
print_tee( time.ctime() )

failed_tests = []
passed_tests = []
ntests = len(opts.tests)
run_all = (ntests == 0)

seen = set()
for cmd in cmds:
    if ((run_all or cmd[0] in opts.tests) and cmd[0] not in opts.exclude):
        if (start_routine and cmd[0] != start_routine):
            print_tee( 'skipping', cmd[0] )
            continue
        start_routine = None

        seen.add( cmd[0] )
        (err, output) = run_test( cmd )
        if (err):
            failed_tests.append( (cmd[0], err, output) )
        else:
            passed_tests.append( cmd[0] )

not_seen = list( filter( lambda x: x not in seen, opts.tests ) )
if (not_seen):
    print_tee( 'Warning: unknown routines:', ' '.join( not_seen ))

# print summary of failures
nfailed = len( failed_tests )
if (nfailed > 0):
    print_tee( '\n' + str(nfailed) + ' routines FAILED:',
               ', '.join( [x[0] for x in failed_tests] ) )
else:
    print_tee( '\n' + 'All routines passed.' )

# generate jUnit compatible test report
if opts.xml:
    print( 'writing XML file', opts.xml )
    root = ET.Element("testsuites")
    doc = ET.SubElement(root, "testsuite",
                        name="lapackpp_suite",
                        tests=str(ntests),
                        errors="0",
                        failures=str(nfailed))

    for (test, err, output) in failed_tests:
        testcase = ET.SubElement(doc, "testcase", name=test)

        failure = ET.SubElement(testcase, "failure")
        if (err < 0):
            failure.text = "exit with signal " + str(-err)
        else:
            failure.text = str(err) + " tests failed"

        system_out = ET.SubElement(testcase, "system-out")
        system_out.text = output
    # end

    for test in passed_tests:
        testcase = ET.SubElement(doc, 'testcase', name=test)
        testcase.text = 'PASSED'

    tree = ET.ElementTree(root)
    indent_xml( root )
    tree.write( opts.xml )
# end

elapsed = time.time() - start
print_tee( 'Elapsed %.2f sec' % elapsed )
print_tee( time.ctime() )

exit( nfailed )



# Copyright (c) 2017-2023, University of Tennessee. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# This program is free software: you can redistribute it and/or modify it under
# the terms of the BSD 3-Clause license. See the accompanying LICENSE file.

from __future__ import print_function

import os
import shlex
import subprocess
from   subprocess import PIPE
import math
import sys
import time
import re
import tarfile
import argparse
import shutil

# This relative import syntax works in both python2 and 3.
from .ansicodes import font

# Python 3 renames raw_input => input.
if (sys.version_info.major < 3):
    input = raw_input

#-------------------------------------------------------------------------------
def urlretrieve( url, filename ):
    '''
    Downloads url and saves to filename.
    Works for both Python 2 and 3, which differ in where urlretrieve is located.
    '''
    if (sys.version_info.major >= 3):
        import urllib.request as urllib_request
    else:
        import urllib as urllib_request
    urllib_request.urlretrieve( url, filename )
# end

#-------------------------------------------------------------------------------
interactive_ = False

# Function to get and set interactive flag. If True, config finds all possible
# values and gives user a choice. If False, config picks the first valid value.
# value = interactive() returns value of interactive.
# interactive( value ) sets value of interactive.
# Making this a function config.interactive() avoids issues with the
# package __init__.py importing it if it were a variable.
def interactive( value=None ):
    global interactive_
    if (value is not None):
        interactive_ = value
    return interactive_
# end

# ------------------------------------------------------------------------------
debug_ = False

def debug( value=None ):
    global debug_
    if (value is not None):
        debug_ = value
    return debug_
# end

# ------------------------------------------------------------------------------
namespace_ = None

def namespace( value ):
    return namespace_

def define( var, value=None ):
    txt = '-D' + namespace_ + '_' + var
    if (value):
        txt += '=' + value
    return txt

# ------------------------------------------------------------------------------
# variables to replace instead of appending/prepending
replace_vars = ['CC', 'CXX', 'NVCC', 'FC', 'AR', 'RANLIB', 'prefix']

# ------------------------------------------------------------------------------
# map file extensions to languages
lang_map = {
    '.c':   'CC',

    '.cc':  'CXX',
    '.cxx': 'CXX',
    '.cpp': 'CXX',

    '.cu':  'NVCC',

    '.f':   'FC',
    '.f90': 'FC',
    '.f77': 'FC',
    '.F90': 'FC',
    '.F77': 'FC',
}

# ------------------------------------------------------------------------------
# map languages to compiler flags
flag_map = {
    'CC':   'CFLAGS',
    'CXX':  'CXXFLAGS',
    'NVCC': 'NVCCFLAGS',
    'FC':   'FFLAGS',
}

# ------------------------------------------------------------------------------
def flatten( data, ltypes=(list, tuple) ):
    '''
    Flattens nested list or tuple.
    Ex: flatten( [1, 2, [3, [4, 5], 6]] ) returns [1, 2, 3, 4, 5, 6]

    see http://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html
    '''
    ltype = type(data)
    data = list(data)
    i = 0
    while i < len(data):
        while isinstance(data[i], ltypes):
            if not data[i]:
                data.pop(i)
                i -= 1
                break
            else:
                data[i:i + 1] = data[i]
        i += 1
    return ltype(data)
# end

#-------------------------------------------------------------------------------
def get( dictionary, key ):
    '''
    Returns dictionary[ key ] or ''
    '''
    if (key in dictionary):
        return dictionary[ key ]
    else:
        return ''
# end

#-------------------------------------------------------------------------------
def print_header( header ):
    '''
    Prints a header, with bold font, both to console and the log.
    '''
    txt = font.bold( header )
    print( '\n' + '-'*80 +
           '\n' + txt, file=log )
    print( '\n' + txt )
# end

#-------------------------------------------------------------------------------
def print_subhead( subhead ):
    '''
    Prints a subhead, both to console and the log.
    '''
    print( '-'*40 + '\n' +
           subhead, file=log )
    print( subhead )
# end

#-------------------------------------------------------------------------------
def print_msg( msg ):
    '''
    Prints msg, both to console and the log.
    '''
    print( msg, file=log )
    print( msg )
# end

#-------------------------------------------------------------------------------
def print_warn( msg ):
    '''
    Prints warning msg, with bold red font, both to console and the log.
    '''
    txt = font.bold( font.red( 'Warning: ' + msg ) )
    print( txt, file=log )
    print( txt )
# end

#-------------------------------------------------------------------------------
def print_test( label ):
    '''
    If label is given, prints the label, both to console and the log.
    On the console, it doesn't print the trailing newline; a subsequent
    print_result() will print it.
    If no label is given, does nothing. This simplifies functions like
    compile_obj that take an optional label to print.
    '''
    if (label):
        print( '-'*20 + '\n' + label, file=log )
        print( '%-72s' % label, end='' )
        sys.stdout.flush()
# end

#-------------------------------------------------------------------------------
def print_result( label, rc, extra='' ):
    '''
    If label is given, prints either "yes" (if rc == 0) or "no" (otherwise).
    Extra is printed after yes or no.
    If no label is given, does nothing.
    @see print_test().
    '''
    if (label):
        if (rc == 0):
            print( font.blue( 'yes'  ), extra, file=log )
            print( font.blue( ' yes' ), extra )
        else:
            print( font.red( 'no'  ), extra, file=log )
            print( font.red( ' no' ), extra )
# end

# ------------------------------------------------------------------------------
# Used for all errors.
# Allows Python Exceptions to fall through, giving tracebacks.
class Error( Exception ):
    pass

class Quit( Error ):
    pass

#-------------------------------------------------------------------------------
class Environments( object ):
    '''
    Manages stack of environments, which are dictionaries of name=value pairs.
    '''

    # ----------------------------------------
    def __init__( self ):
        '''
        Initializes the environment stack.
        The bottom is os.environ. The top is an empty environment.
        '''
        self.stack = [ os.environ, {} ]

    # ----------------------------------------
    def push( self, env=None ):
        '''
        Push an empty enviroment on the environment stack.
        If env is given, also merge env into the environment stack.
        '''
        self.stack.append( {} )
        if (env):
            self.merge( env )

    # ----------------------------------------
    def top( self ):
        '''
        Return top-most environment in the environment stack.
        '''
        return self.stack[-1]

    # ----------------------------------------
    def pop( self ):
        '''
        Remove the top-most environment from the environment stack.
        '''
        if (len(self.stack) == 2):
            raise Error( "can't pop last 2 environments" )
        return self.stack.pop()

    # ----------------------------------------
    def __contains__( self, key ):
        '''
        Returns true if a key exists in the environment stack.
        '''
        for env in self.stack[::-1]:
            if (key in env):
                return True
        return False

    # ----------------------------------------
    def __getitem__( self, key ):
        '''
        Returns the value of the key, searching from the top of the environment
        stack down. As in a Makefile, unknown keys return empty string ('').
        Use 'x in environ' to test whether a key exists.
        '''
        for env in self.stack[::-1]:
            if (key in env):
                return env[ key ]
        return ''

    # ----------------------------------------
    def __setitem__( self, key, value ):
        '''
        Sets the key's value
        in the top-most environment in the environment stack.
        '''
        self.stack[ -1 ][ key ] = value

    # ----------------------------------------
    def append( self, key, val ):
        '''
        Append val to key's value, saving the result
        in the top-most environment in the enviornment stack.
        '''
        orig = self[ key ]
        if (val):
            if (orig):
                val = orig + ' ' + val
            self[ key ] = val
        return orig

    # ----------------------------------------
    def prepend( self, key, val ):
        '''
        Prepend val to key's value, saving the result
        in the top-most environment in the enviornment stack.
        '''
        orig = self[ key ]
        if (val):
            if (orig):
                val = val + ' ' + orig
            self[ key ] = val
        return orig

    # ----------------------------------------
    def merge( self, env ):
        '''
        Merges env, a dictionary of environment variables, into the existing
        environment stack. For most variables, the value in env is appended
        to any existing value. For LIBS, the value is prepended.
        For variables in config.replace_vars (like CXX), the value in env
        replaces the existing value.
        '''
        for key in env:
            if (key in replace_vars):
                self[ key ] = env[ key ]
            elif (key == 'LIBS'):
                self.prepend( key, env[ key ] )
            else:
                self.append( key, env[ key ] )
# end

#-------------------------------------------------------------------------------
def choose( prompt, choices ):
    '''
    Asks the user to choose among the given choices.
    Returns the index of the chosen item in the range [0, len(choices)-1],
    or raises Error or Quit exceptions.
    '''
    choices = list( choices )
    n = len( choices )
    if (n == 0):
        print( font.bold( font.red( 'none found' ) ) )
        raise Error
    elif (n == 1):
        ##print()
        return 0
    else:
        width = int( math.log10( n ) + 1 )
        print( '\n' + prompt )
        for i in range( n ):
            print( '[%*d] %s' % (width, i+1, choices[i]) )
        while (True):
            print( 'Enter [1-%d] or quit: ' % (n), end='' )
            sys.stdout.flush()
            i = input()
            if (i == 'q' or i == 'quit'):
                raise Quit
            try:
                i = int( i )
            except:
                i = -1
            if (i >= 1 and i <= len( choices )):
                ##print()
                return i-1
        # end
    # end
# end

#-------------------------------------------------------------------------------
def run( cmd, env=None ):
    '''
    Runs the command cmd.
    cmd can be a string or a nested list.
    Pushes env beforehand and pops afterward.
    stdout and stderr are written to the log.
    Returns (return_code, stdout, stderr) from the command.

    Ex: run( ['gcc', '-c', 'file.c'], {'CPATH': '/opt/include'} )
    runs: gcc -c file.c
    '''
    environ.push( env )

    if (not isinstance( cmd, str )):
        cmd = ' '.join( flatten( cmd ))

    print( '>>>', cmd, file=log )
    cmd_list = shlex.split( cmd )
    try:
        proc = subprocess.Popen( cmd_list, stdout=PIPE, stderr=PIPE )
        (stdout, stderr) = proc.communicate()
        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')

        rc = proc.wait()
        log.write( stdout )
        if (stderr):
            log.write( font.red( stderr ) )
        print( 'exit status = %d' % rc, file=log )
    except Exception as ex:
        print( 'Exception:', str(ex), file=log )
        rc = -1
        stdout = ''
        stderr = str(ex)

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
def compile_obj( src, env=None, label=None ):
    '''
    Compiles source file src into an object (.o) file.
    Pushes env beforehand and pops afterwards.
    If label is given, prints label & result.
    Returns (return_code, stdout, stderr) from the compiler.

    Ex: compile_obj( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
    runs: gcc $CFLAGS -c foo.c -o foo.o
    '''
    environ.push( env )

    print_test( label )
    (base, ext) = os.path.splitext( src )
    obj      = base + '.o'
    lang     = lang_map[ ext ]
    compiler = environ[ lang ]
    flags    = environ[ flag_map[ lang ]]
    (rc, stdout, stderr) = run([ compiler, flags, '-c', src, '-o', obj ])
    print_result( label, rc )

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
def link_exe( src, env=None, label=None ):
    '''
    Links the object file (.o) associated with the source file src into an executable.
    Assumes compile_obj( src ) was called previously to generate the object file.
    Pushes env beforehand and pops afterward.
    If label is given, prints label & result.
    Returns (return_code, stdout, stderr) from the compiler.

    Ex: link_exe( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
    runs: gcc $LDFLAGS $LIBS foo.o -o foo
    '''
    environ.push( env )

    print_test( label )
    (base, ext) = os.path.splitext( src )
    obj      = base + '.o'
    lang     = lang_map[ ext ]
    compiler = environ[ lang ]
    LDFLAGS  = environ['LDFLAGS']
    LIBS     = environ['LIBS'] or environ['LDLIBS']
    (rc, stdout, stderr) = run([ compiler, obj, '-o', base, LDFLAGS, LIBS ])
    print_result( label, rc )

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
def compile_exe( src, env=None, label=None ):
    '''
    Compiles source file src into an object file via compile_obj(),
    then links it into an exe.

    Ex: compile_exe( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
    runs: gcc $CFLAGS -c foo.c -o foo.o
          gcc $LDFLAGS $LIBS foo.o -o foo
    '''
    environ.push( env )

    print_test( label )
    (base, ext) = os.path.splitext( src )
    obj      = base + '.o'
    lang     = lang_map[ ext ]
    compiler = environ[ lang ]
    LDFLAGS  = environ['LDFLAGS']
    LIBS     = environ['LIBS'] or environ['LDLIBS']
    (rc, stdout, stderr) = compile_obj( src )
    if (rc == 0):
        (rc, stdout, stderr) = run([ compiler, obj, '-o', base, LDFLAGS, LIBS ])
    print_result( label, rc )

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
# Ex:
# compile_run( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
def compile_run( src, env=None, label=None ):
    '''
    Compiles source file src into an object file and exe via compile_exe(),
    then executes the exe.

    Ex: compile_exe( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
    runs: gcc $CFLAGS -c foo.c -o foo.o
          gcc $LDFLAGS $LIBS foo.o -o foo
          ./foo
    '''
    environ.push( env )

    print_test( label )
    (base, ext) = os.path.splitext( src )
    (rc, stdout, stderr) = compile_exe( src )
    if (rc == 0):
        (rc, stdout, stderr) = run( './' + base )
    print_result( label, rc )

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
def run_exe( src, env=None, label=None ):
    '''
    Runs the exe associated with src.
    Assumes compile_exe( src ) was called previously to generate the exe.

    Ex: run_exe( 'foo.c', {'CC': 'gcc'}, 'Test foo' )
    runs: ./foo
    '''
    environ.push( env )

    print_test( label )
    (base, ext) = os.path.splitext( src )
    (rc, stdout, stderr) = run( './' + base )
    print_result( label, rc )

    environ.pop()
    return (rc, stdout, stderr)
# end

#-------------------------------------------------------------------------------
def prog_cxx( choices=['g++', 'c++', 'CC', 'cxx', 'icpc', 'xlc++', 'clang++'] ):
    '''
    Searches for available C++ compilers from the list of choices.
    Sets CXX to the chosen one.
    '''
    print_header( 'C++ compiler' )

    cxx = environ['CXX']
    if (cxx):
        print( 'Trying $CXX =', cxx )
        choices = [ cxx ]

    passed = []  # CXX compilers, e.g., g++ or mpicxx
    actual = []  # Detected underlying compilers, e.g., g++ or clang++
    for cxx in choices:
        print_test( cxx )
        (rc, out, err) = compile_run( 'config/compiler_cxx.cc', {'CXX': cxx} )
        # print (g++), (clang++), etc., as output by compiler_cxx, after yes
        if (rc == 0):
            cxx_actual = out.strip()
            out = '(' + cxx_actual + ')'
            actual.append( cxx_actual )
        print_result( cxx, rc, out )
        if (rc == 0):
            passed.append( cxx )
            if (not interactive()): break
        # end
    # end
    i = choose( 'Choose C++ compiler:', passed )
    environ['CXX'] = passed[i]
    environ['CXX_actual'] = actual[i]
# end

#-------------------------------------------------------------------------------
def prog_cxx_flag( flags ):
    '''
    Tests each flag in flags; the first that passes is added to CXXFLAGS.
    flags can be an individual string or an iterable (list, tuple, etc.).
    '''
    if (type( flags ) == str):
        flags = [ flags ]
    # end
    for flag in flags:
        print_test( flag )
        (rc, out, err) = compile_obj( 'config/compiler_cxx.cc', {'CXXFLAGS': flag} )
        # assume a mention of the flag in stderr means it isn't supported
        if (flag in err):
            rc = 1
        print_result( flag, rc )
        if (rc == 0):
            environ.append( 'CXXFLAGS', flag )
            break
    # end
# end

#-------------------------------------------------------------------------------
def openmp( flags=['-fopenmp', '-qopenmp', '-openmp', '-omp', ''] ):
    '''
    Tests for OpenMP support with one of the given flags.
    If a flag works, it is added to both CXXFLAGS and LDFLAGS.
    '''
    print_header( 'OpenMP support' )
    src = 'config/openmp.cc'
    for flag in flags:
        print_test( flag )
        env = {'CXXFLAGS': flag, 'LDFLAGS': flag, 'HAS_OPENMP': True}
        (rc, out, err) = compile_run( src, env )
        print_result( flag, rc )
        if (rc == 0):
            environ.merge( env )
            break
    # end
# end

#-------------------------------------------------------------------------------
def cublas_library():
    '''
    Tests for linking CUDA and cuBLAS libraries.
    Does not actually run the resulting exe, to allow compiling with CUDA on a
    machine without GPUs.
    '''
    # Find CUDA to add -I, -L, -rpath flags.
    # CUDA_PATH is used in NVIDIA Getting Started documentation;
    # CUDA_HOME is used in Spack CUDA package;
    # else infer from `which nvcc`.
    cuda_path = environ['CUDA_PATH']
    if (not cuda_path):
        cuda_path = environ['CUDA_HOME']
    if (not cuda_path):
        nvcc_path = shutil.which( 'nvcc' )
        if (nvcc_path):
            (bin_path, nvcc) = os.path.split( nvcc_path )
            (cuda_path, bin_) = os.path.split( bin_path )

    cxxflags = define('HAVE_CUBLAS')
    ldflags  = ''
    libs     = '-lcusolver -lcublas -lcudart'

    if (cuda_path):
        incdir = os.path.join( cuda_path, 'include' )
        if (os.path.exists( incdir )):
            cxxflags += ' -I' + incdir

        libdir = os.path.join( cuda_path, 'lib64' )
        if (not os.path.exists( libdir )):
            libdir = os.path.join( cuda_path, 'lib' )
        if (os.path.exists( libdir )):
            ldflags += '-L' + libdir + ' -Wl,-rpath,' + libdir
    # end

    print_subhead( 'CUDA and cuBLAS libraries' )
    print_test( '    ' + cxxflags + ' ' + ldflags + ' ' + libs )
    env = {'CXXFLAGS': cxxflags, 'LDFLAGS': ldflags, 'LIBS': libs}
    (rc, out, err) = compile_exe( 'config/cublas.cc', env )
    print_result( libs, rc )
    if (rc == 0):
        environ.merge( env )
    else:
        raise Error( 'cuBLAS not found' )
# end

#-------------------------------------------------------------------------------
def rocblas_library():
    '''
    Tests for linking ROCm/HIP and rocBLAS libraries.
    Does not actually run the resulting exe, to allow compiling with ROCm on a
    machine without GPUs.
    '''
    # Find ROCm to add -I, -L, -rpath flags.
    # ROCM_PATH is used in hipcc and Spack ROCm package;
    # else infer from `which hipcc`.
    rocm_path = environ['ROCM_PATH']
    if (not rocm_path):
        hipcc_path = shutil.which( 'hipcc' )
        if (hipcc_path):
            (bin_path, hipcc) = os.path.split( hipcc_path )
            (rocm_path, bin_) = os.path.split( bin_path )

    cxxflags = define('HAVE_ROCBLAS')
    ldflags  = ''
    libs     = '-lrocsolver -lrocblas -lamdhip64'

    if (rocm_path):
        incdir = os.path.join( rocm_path, 'include' )
        if (os.path.exists( incdir )):
            cxxflags += ' -I' + incdir

        # Some versions of ROCm (5.1.3) have both lib and lib64 directories;
        # we need the lib directory.
        libdir = os.path.join( rocm_path, 'lib' )
        if (not os.path.exists( libdir )):
            libdir = os.path.join( rocm_path, 'lib64' )
        if (os.path.exists( libdir )):
            ldflags += ' -L' + libdir + ' -Wl,-rpath,' + libdir
    # end

    print_subhead( 'HIP/ROCm and rocBLAS libraries' )
    print_test( '    ' + cxxflags + ' ' + ldflags + ' ' + libs )
    env = {'CXXFLAGS': cxxflags, 'LDFLAGS': ldflags, 'LIBS': libs}
    (rc, out, err) = compile_exe( 'config/rocblas.cc', env )
    print_result( libs, rc )
    if (rc == 0):
        environ.merge( env )
    else:
        raise Error( 'rocBLAS not found' )
# end

#-------------------------------------------------------------------------------
def sycl_onemkl_library():
    '''
    Tests for linking SYCL and oneMKL library.
    Does not actually run the resulting exe, to allow compiling on a
    machine without GPUs.
    '''
    libs = '-lmkl_sycl -lsycl -lOpenCL'
    print_subhead( 'SYCL and oneMKL libraries' )
    print_test( '    ' + libs )

    # Intel compiler vars.sh defines $CMPLR_ROOT
    root = environ['CMPLR_ROOT'] or environ['CMPROOT']
    inc = ''
    if (root):
        inc = '-I' + root + '/linux/include ' \
            + '-I' + root + '/linux/include/sycl '
    env = {'LIBS': libs,
           'CXXFLAGS': inc + define('HAVE_SYCL')
           + ' -fsycl -fp-model=precise' }
    (rc, out, err) = compile_exe( 'config/onemkl.cc', env )
    print_result( libs, rc )
    if (rc == 0):
        environ.merge( env )
    else:
        raise Error( 'oneMKL not found' )
# end

#-------------------------------------------------------------------------------
def gpu_blas():
    gpu_backend = environ['gpu_backend'] or 'auto'
    print_header( 'GPU BLAS libraries: gpu_backend = ' + gpu_backend )

    test_auto   = re.search( r'\b(auto)\b',     gpu_backend )
    test_cuda   = re.search( r'\b(cuda)\b',     gpu_backend ) or test_auto
    test_rocm   = re.search( r'\b(hip|rocm)\b', gpu_backend ) or test_auto
    test_sycl   = re.search( r'\b(sycl)\b',     gpu_backend ) or test_auto

    #----- CUDA
    gpu_blas_found = False
    if (test_cuda):
        try:
            cublas_library()
            gpu_blas_found = True
        except Error as ex:
            if (gpu_backend == 'cuda'):
                raise ex  # fatal
    else:
        print_msg( font.red( 'skipping CUDA search' ) )

    #----- ROCm
    if (not gpu_blas_found and test_rocm):
        try:
            rocblas_library()
            gpu_blas_found = True
        except Error as ex:
            if (gpu_backend in ('hip', 'rocm')):
                raise ex  # fatal
    else:
        print_msg( font.red( 'skipping HIP/ROCm search' ) )

    #----- SYCL
    if (not gpu_blas_found and test_sycl):
        try:
            sycl_onemkl_library()
            gpu_blas_found = True
        except Error as ex:
            if (gpu_backend == 'sycl'):
                raise ex  # fatal
    else:
        print_msg( font.red( 'skipping SYCL search' ) )

    if (not gpu_blas_found):
        print_warn( 'No GPU BLAS library found' )
# end

#-------------------------------------------------------------------------------
def get_package( name, directories, unique_file, repo_url, tar_url, tar_filename ):
    '''
    Searches for a package, generally used for internal packages.
    Looks for a directory from given list of directories. It checks for
    existence of unique_file in the directory, to ensure the directory
    actually contains the source.
    If found, returns that directory.
    If none found, tries to 'git clone repo_url' to the last directory.
    If that fails, tries to download tar_url and unpack it to the last directory.
    '''
    global log

    print_header( name )

    for directory in directories:
        path = os.path.join( directory, unique_file )
        print_test( path )
        err = not os.path.exists( path )
        print_result( directory, err )
        if (not err):
            return directory
    # end

    if (repo_url):
        if (interactive()):
            print( name +' not found; git clone '+ repo_url +'? [Y/n] ', end='' )
            sys.stdout.flush()
            i = input().lower()
        if (not interactive() or i in ('', 'y', 'yes')):
            cmd = 'git clone '+ repo_url +' '+ directory
            print_test( 'download: ' + cmd )
            (err, stdout, stderr) = run( cmd )
            print_result( 'download', err )
            if (not err):
                return directory
    # end

    if (tar_url):
        if (interactive()):
            print( name +' not found; download from '+ tar_url +'? [Y/n] ', end='' )
            sys.stdout.flush()
            i = input().lower()
        if (not interactive() or i in ('', 'y', 'yes')):
            try:
                print_test( 'download: '+ tar_url +' as '+ tar_filename )
                urlretrieve( tar_url, tar_filename )

                print( 'untar', tar_filename, file=log )
                tar = tarfile.open( tar_filename )
                files = tar.getnames()
                last = ''
                for f in files:
                    # sanitize file names: disallow beginning with / or having ../
                    if (re.search( r'^/|\.\./', f )):
                        print( 'skipping', f )
                        continue
                    tar.extract( f )
                    lastfile = f
                # end

                # rename directory,
                # e.g., from icl-testsweeper-dbd960ebf706 to testsweeper
                # todo: os.path.sep intsead of '/'?
                dirs = re.split( '/', lastfile )
                print( 'rename', dirs[0], directory, file=log )
                os.rename( dirs[0], directory )
                err = 0
            except Exception as ex:
                print( 'Exception:', str(ex), file=log )
            # end
            print_result( 'download', err )
            if (not err):
                return directory
        # end
    # end

    # otherwise, not found
    return None
# end

#-------------------------------------------------------------------------------
def extract_defines_from_flags( flags='CXXFLAGS', var='HEADER_DEFINES' ):
    '''
    Extracts all "-Dname[=value]" defines from the given flags.
    Adds all "-Dname[=value]" defines to DEFINES.
    Adds all "#define name [value]" defines to HEADER_DEFINES.
    Stores all name=value defines for autoconf-like "#undef name" substitution
    in output_files().
    '''
    global environ, defines
    exp = r'(-D(\w+)(?:=(\S*))?) *'
    defs = re.findall( exp, environ[ flags ] )
    environ[ flags ] = re.sub( exp, '', environ[ flags ] ).strip()
    header = ''
    for (name_value, name, value) in defs:
        environ.append( 'DEFINES', name_value )
        defines[ name ] = value
        if (value):
            header += '#define '+ name +' '+ value + '\n'
        else:
            header += '#define '+ name + '\n'
    # end
    environ[ var ] = header
# end

#-------------------------------------------------------------------------------
def sub_env( match ):
    '''
    Given a re (regular expression) match object, returns value of environment variable.
    Used in output_files().
    '''
    return environ[ match.group(1) ]

#-------------------------------------------------------------------------------
def sub_define( match ):
    '''
    Given a re regexp match object,
    returns "#define name [value]" or "// #undef name"
    Used in output_files().
    '''
    global defines
    name = match.group(1)
    if (name in defines):
        value = defines[ name ]
        if (value):
            return '#define '+ name +' '+ value
        else:
            return '#define '+ name
    else:
        return '// #undef '+ name
# end

#-------------------------------------------------------------------------------
def read( filename ):
    '''
    Reads and returns the entire contents of filename.
    '''
    f = open( filename, 'r' )
    txt = f.read()
    f.close()
    return txt
# end

#-------------------------------------------------------------------------------
def write( filename, txt ):
    '''
    Writes txt to filename.
    '''
    f = open( filename, 'w' )
    f.write( txt )
    f.close()
# end

#-------------------------------------------------------------------------------
def output_files( files ):
    '''
    Create each file in files from file.in, substituting @foo@ with variable foo.
    This avoids re-creating the file if the contents did not change.
    files can be a single file or list of files.
    '''
    print_header( 'Output files' )
    if (isinstance( files, str )):
        files = [ files ]
    for fname in files:
        txt = read( fname + '.in' )
        txt = re.sub( r'@(\w+)@', sub_env, txt )
        txt = re.sub( r'#undef (\w+)', sub_define, txt )
        exists = os.path.exists( fname )
        if (exists and txt == read( fname )):
            print( fname, 'is unchanged' )
        else:
            if (exists):
                bak = fname + '.bak'
                print( 'backing up', fname, 'to', bak )
                os.rename( fname, bak )
            # end
            print( 'creating', fname )
            write( fname, txt )
        # end
    # end
# end

#-------------------------------------------------------------------------------
def parse_args():
    '''
    Parses command line options.
    Sets if interactive and if ansicodes are enabled.
    '''
    global opts, parser, debug

    #--------------------
    # Parse command line. We'll handle help ourselves.
    parser = argparse.ArgumentParser( add_help=False )
    parser.add_argument( '-i', '--interactive', action='store_true',
                         help='Find all available choices and ask user which to use;'
                             +' otherwise use first choice found.' )
    parser.add_argument( '--color', action='store', default='auto',
                         help='Use ANSI colors: yes, no, or auto; default %(default)s.' )
    parser.add_argument( '--debug', action='store_true',
                         help='Enable debugging output.' )
    parser.add_argument( '-h', '--help', action='store_true',
                         help='Print help and exit.' )
    parser.add_argument( 'options', nargs=argparse.REMAINDER,
                         help='name=value pairs of options to define.' )
    opts = parser.parse_args()

    # Parse name=value pairs.
    for arg in opts.options:
        s = re.search( '^(\w+)=(.*)', arg )
        if (s):
            environ[ s.group(1) ] = s.group(2)
        else:
            print( 'Unknown argument:', arg )
            exit(1)
    # end

    if (environ['color']):
        opts.color = environ['color']
    font.set_enabled( opts.color )

    if (environ['interactive']):
        opts.interactive = environ['interactive']
    if (opts.interactive):
        interactive( True )

    debug( opts.debug )
# end

#-------------------------------------------------------------------------------
def init( namespace, prefix='/usr/local' ):
    '''
    Initializes config.
    Opens the logfile and deals with OS-specific issues.
    '''
    global log, namespace_

    namespace_ = namespace

    # Default prefix.
    if (not environ['prefix']):
        environ['prefix'] = prefix

    #--------------------
    logfile = 'config/log.txt'
    print( 'opening log file ' + logfile + '\n' )
    log = open( logfile, 'w' )

    #--------------------
    # Workaround if MacOS SIP may have prevented inheriting DYLD_LIBRARY_PATH.
    if (sys.platform.startswith('darwin') and
        'LD_LIBRARY_PATH' not in os.environ and
        'DYLD_LIBRARY_PATH' not in os.environ):

        txt = font.bold( 'NOTICE: $DYLD_LIBRARY_PATH was not set or not inherited.\n' )
        if ('LIBRARY_PATH' in os.environ):
            os.environ['DYLD_LIBRARY_PATH'] = os.environ['LIBRARY_PATH']
            txt += 'Setting $DYLD_LIBRARY_PATH = $LIBRARY_PATH to run test programs.\n'
            txt += '$LIBRARY_PATH = ' + os.environ['LIBRARY_PATH'] + '\n'
        else:
            txt += '$LIBRARY_PATH is not set. Leaving $DYLD_LIBRARY_PATH unset.\n'
        # end
        txt += '''
MacOS System Integrity Protection (SIP) prevents configure.py from inheriting
$DYLD_LIBRARY_PATH. Using
    python3 configure.py
directly (not via make), with a 3rd party python3 from python.org, Homebrew, etc.
(i.e., not /usr/bin/python3), will allow $DYLD_LIBRARY_PATH to be inherited.
'''
        txt = font.red( txt )
        txt += '-'*80
        print( txt )
        print( txt, file=log )
    # end

    #--------------------
    if (opts.help):
        parser.print_help()
        exit(0)
# end

# ------------------------------------------------------------------------------
# Initialize global variables here, rather than in init(),
# so they are imported by __init__.py.
environ = Environments()
environ['argv'] = ' '.join( sys.argv )
environ['datetime'] = time.ctime()

defines = {}

# Parse command line early, so ANSI codes are enabled or disabled early on.
parse_args()



# Extracts initial substring delimited by characters start and end.
# Returns substring, excluding delimiters, and remainder. Example:
# (extracted, remainder) = extract_bracketed( "(n*(n+1)) other", '(', ')' )
# extracted = "n*(n+1)"
# remainder = " other"
# see https://stackoverflow.com/questions/1651487/python-parsing-bracketed-blocks
def extract_bracketed( txt, start, end ):
    if (txt[0] != start):
        return (None, txt)
    cnt = 0
    for i in range( 0, len( txt )):
        if (txt[i] == start):
            cnt += 1
        elif (txt[i] == end):
            cnt -= 1
        if (cnt == 0):
            return (txt[1:i], txt[i+1:])
    # unbalanced
    return (None, txt)
# end
